from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, Generic, Optional, Protocol, Sequence, Type, TypeVar, Union, cast

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from chalk.features._encoding.json import (
    FeatureEncodingOptions,
    structs_as_objects_feature_json_converter,
    structure_json_to_primitive,
    unstructure_primitive_to_json,
)
from chalk.features._encoding.missing_value import MissingValueStrategy
from chalk.features._encoding.primitive import TPrimitive
from chalk.features._encoding.pyarrow import pyarrow_to_polars, pyarrow_to_primitive, rich_to_pyarrow
from chalk.features._encoding.rich import structure_primitive_to_rich, unstructure_rich_to_primitive
from chalk.parsed.proto import data_pb2 as pb
from chalk.utils.collections import unwrap_annotated_if_needed
from chalk.utils.json import TJSON
from chalk.utils.log_with_context import get_logger

_logger = get_logger(__name__)

if TYPE_CHECKING:
    import polars as pl

_TRich = TypeVar("_TRich")
_TRichCo = TypeVar("_TRichCo", covariant=True)
_TRichCon = TypeVar("_TRichCon", contravariant=True)

_TPrim = TypeVar("_TPrim", bound=TPrimitive)
_TPrimCo = TypeVar("_TPrimCo", bound=TPrimitive, covariant=True)
_TPrimCon = TypeVar("_TPrimCon", bound=TPrimitive, contravariant=True)

_DEFAULT_FEATURE_ENCODING_OPTIONS = FeatureEncodingOptions()


def _recursively_unwrap_numpy(x: Any) -> Any:
    if isinstance(x, list):
        return [_recursively_unwrap_numpy(y) for y in x]
    if isinstance(x, dict):
        return {k: _recursively_unwrap_numpy(v) for (k, v) in x.items()}
    if isinstance(x, np.number):
        return x.item()
    return x


class MissingValueError(TypeError):
    """Raised when a missing value is encountered and the missing value strategy is set to ``error``."""

    pass


class TEncoder(Protocol[_TPrimCo, _TRichCon]):
    def __call__(self, value: _TRichCon, /) -> _TPrimCo:
        ...


class TDecoder(Protocol[_TPrimCon, _TRichCo]):
    def __call__(self, value: _TPrimCon, /) -> _TRichCo:
        ...


class FeatureConverter(Generic[_TPrim, _TRich]):
    def __init__(
        self,
        name: str,
        is_nullable: bool,
        rich_default: Union[_TRich, ellipsis] = ...,
        primitive_default: Union[_TPrim, ellipsis] = ...,
        rich_type: Union[Type[_TRich], ellipsis] = ...,
        pyarrow_dtype: Optional[pa.DataType[Any]] = None,
        encoder: Optional[TEncoder[_TPrim, _TRich]] = None,
        decoder: Optional[TDecoder[_TPrim, _TRich]] = None,
    ) -> None:
        super().__init__()
        self._name = name
        self._rich_type = unwrap_annotated_if_needed(rich_type)
        if pyarrow_dtype is None:
            if rich_type is ...:
                raise ValueError("Either the `rich_type` or `pyarrow_dtype` must be provided")
            pyarrow_dtype = rich_to_pyarrow(rich_type, name)

        if rich_type is ...:
            if rich_default != ...:
                raise ValueError(
                    "The `rich_default` cannot be used without the `rich_type`. Perhaps specify the `primitive_default` instead?"
                )
            if is_nullable and primitive_default is ...:
                primitive_default = cast(_TPrim, None)

        else:
            if primitive_default != ...:
                raise ValueError(
                    "The `primitive_default` cannot be used when specifying the `rich_type`. Instead, specify the `rich_default`."
                )
            if is_nullable and rich_default is ...:
                rich_default = cast(_TRich, None)

        self._pyarrow_dtype = pyarrow_dtype
        if pa.types.is_struct(pyarrow_dtype) and is_nullable:
            is_nullable = False
        self._is_nullable = is_nullable
        self._primitive_type = pyarrow_to_primitive(self._pyarrow_dtype, name)
        if rich_type is ...:
            if encoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
            if decoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
        self._encoder = encoder
        self._decoder = decoder
        self._rich_default = rich_default
        if primitive_default is ... and rich_default != ...:
            # The missing value strategy doesn't really matter because rich_default is not missing
            primitive_default = self.from_rich_to_primitive(rich_default, missing_value_strategy="allow")
        self._primitive_default = primitive_default

    @property
    def encoder(self):
        return self._encoder

    @property
    def decoder(self):
        return self._decoder

    def _to_primitive(self, val: _TRich) -> _TPrim:
        if val is None or self._encoder is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                x = unstructure_rich_to_primitive(val)
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}'"
                ) from e
            if x is None and not self._is_nullable:
                raise ValueError(f"Feature '{self._name}' is null, but it cannot be nullable")
            try:
                return cast(_TPrim, structure_primitive_to_rich(x, cast(Type[_TRich], self.primitive_type)))
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}"
                ) from e
        return self._encoder(val)

    def _from_prim(self, val: Union[_TPrim, _TRich]) -> _TRich:
        if self._rich_type is ...:
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        if val is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                val = structure_primitive_to_rich(cast(_TPrim, val), cast(Type[_TRich], self.primitive_type))
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}'"
                ) from e
        if self._decoder is None:
            try:
                return structure_primitive_to_rich(cast(_TPrim, val), self._rich_type)
            except (TypeError, ValueError) as e:
                raise TypeError(f"Could not convert '{val}' to `{self._rich_type}` for feature '{self._name}'") from e
        if isinstance(val, self._rich_type):
            return cast(_TRich, val)
        return self._decoder(cast(_TPrim, val))

    def from_rich_to_pyarrow(
        self,
        values: Sequence[Union[_TRich, ellipsis, None]],
        /,
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> Union[pa.Array, pa.ChunkedArray]:
        prim_values = [self.from_rich_to_primitive(x, missing_value_strategy) for x in values]
        return self.from_primitive_to_pyarrow(prim_values)

    def from_rich_to_primitive(
        self,
        value: Union[_TRich, ellipsis, None],
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> _TPrim:
        # Ensure that the rich value is indeed the rich type
        # For example, if a string is passed in for a datetime value, convert it into a datetime
        if self.is_value_missing(value):
            if missing_value_strategy == "allow":
                return cast(_TPrim, value)
            elif missing_value_strategy in ("default_or_error", "default_or_allow"):
                if self.has_default:
                    return self.primitive_default
                elif missing_value_strategy == "default_or_error":
                    raise TypeError(
                        f"The value for feature '{self._name}' is missing, and this feature has no default value."
                    )
                else:
                    return cast(_TPrim, value)
            elif missing_value_strategy == "error":
                raise MissingValueError(
                    f"The value for feature '{self._name}' is missing, but `replace_missing_with_defaults` was set to False."
                )
            else:
                raise ValueError(
                    (
                        f"Unsupported missing value strategy: {missing_value_strategy}. "
                        "It must be one of 'allow', 'default_or_allow', 'default_or_error', or 'error'."
                    )
                )
        value = self.from_primitive_to_rich(cast(_TPrim, value))
        return self._to_primitive(value)

    def from_rich_to_json(
        self,
        value: Union[_TRich, ellipsis, None],
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
        options: FeatureEncodingOptions = _DEFAULT_FEATURE_ENCODING_OPTIONS,
    ) -> TJSON:
        prim_val = self.from_rich_to_primitive(value, missing_value_strategy)
        return self.from_primitive_to_json(prim_val, options=options)

    def from_pyarrow_to_rich(self, values: Union[pa.Array, pa.ChunkedArray], /) -> Sequence[_TRich]:
        return [self.from_primitive_to_rich(x) for x in self.from_pyarrow_to_primitive(values)]

    def from_pyarrow_to_json(
        self,
        values: Union[pa.Array, pa.ChunkedArray],
        options: FeatureEncodingOptions = _DEFAULT_FEATURE_ENCODING_OPTIONS,
    ) -> Sequence[TJSON]:
        return [self.from_primitive_to_json(x, options=options) for x in self.from_pyarrow_to_primitive(values)]

    def from_pyarrow_to_primitive(self, values: Union[pa.Array, pa.ChunkedArray]) -> Sequence[_TPrim]:
        return [_recursively_unwrap_numpy(x) for x in values.to_pylist()]

    def from_primitive_to_rich(self, value: Union[_TPrim, _TRich]) -> _TRich:
        return self._from_prim(value)

    def from_primitive_to_pyarrow(self, value: Sequence[_TPrim]) -> Union[pa.Array, pa.ChunkedArray]:
        # FIXME: Convert recursively
        if pa.types.is_float16(self._pyarrow_dtype):
            return pa.array(np.array([None if x is ... else x for x in value], np.dtype("float16")))
        if pa.types.is_fixed_size_list(self._pyarrow_dtype):
            assert isinstance(self._pyarrow_dtype, pa.FixedSizeListType)
            list_size: int = self._pyarrow_dtype.list_size  # type: ignore
            value_type: pa.DataType = self._pyarrow_dtype.value_type  # type: ignore
            if pa.types.is_float16(value_type):
                empty = [None] * list_size
                # NOTE: Numpy converts both nan and null to nan for individual array elements.
                # Because pa.compute.if_else doesn't work for float16 arrays, it is difficult to flip the null bits from the python api
                # Since the main use case of FixedSizeList(float16) is for vectors, will leave the nans as-is, because it is more likely
                # for an individual element of a vector to be nan (i.e. overflow from underlying embedding model) than null
                ans = pa.FixedSizeListArray.from_arrays(
                    np.array([empty if x is ... or x is None else x for x in value], np.dtype("float16")).reshape(-1),
                    list_size,
                )
                mask = [x is not ... and x is not None for x in value]
                ans = pc.if_else(pa.array(mask, pa.bool_()), ans, pa.scalar(None, self._pyarrow_dtype))  # type: ignore
                return ans
        return pa.array([None if x is ... else x for x in value], type=self._pyarrow_dtype)

    def from_primitive_to_json(
        self, value: TPrimitive, options: FeatureEncodingOptions = _DEFAULT_FEATURE_ENCODING_OPTIONS
    ) -> TJSON:
        if options.encode_structs_as_objects:
            return structs_as_objects_feature_json_converter.unstructure_primitive_to_json(value)
        else:
            return unstructure_primitive_to_json(value)

    def from_json_to_rich(self, value: TJSON) -> _TRich:
        prim_val = self.from_json_to_primitive(value)
        return self.from_primitive_to_rich(prim_val)

    def from_json_to_pyarrow(self, values: Sequence[TJSON]) -> Union[pa.Array, pa.ChunkedArray]:
        primitive_vals = [self.from_json_to_primitive(x) for x in values]
        return self.from_primitive_to_pyarrow(primitive_vals)

    def from_json_to_primitive(self, value: Union[TJSON, TPrimitive]) -> _TPrim:
        try:
            return cast(_TPrim, structure_json_to_primitive(value, self._primitive_type))
        except (ValueError, TypeError) as e:
            raise TypeError(f"Could not convert '{value}' to `{self._primitive_type}`") from e

    @property
    def pyarrow_dtype(self):
        return self._pyarrow_dtype

    @property
    def is_nullable(self):
        return self._is_nullable

    @property
    def rich_type(self) -> Type[_TRich]:
        if self._rich_type is ...:
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        return self._rich_type

    @property
    def primitive_type(self) -> Type[TPrimitive]:
        return self._primitive_type

    @functools.cached_property
    def polars_dtype(self) -> pl.PolarsDataType:
        return pyarrow_to_polars(self.pyarrow_dtype, self._name)

    @property
    def has_default(self):
        # primitive default is set <==> rich default is also set or no rich type was provided
        return self._primitive_default != ...

    @property
    def rich_default(self) -> _TRich:
        if self._rich_default is ...:
            raise ValueError(f"Feature '{self._name}' has no default value")
        return self._rich_default

    @property
    def primitive_default(self) -> _TPrim:
        if self._primitive_default is ...:
            raise ValueError(f"Feature '{self._name}' has no default value")
        return self._primitive_default

    def is_value_missing(self, value: Any):
        """Returns whether the ``value`` should be treated as a "missing" value"""
        if value is ...:
            # Ellipsis is always missing
            return True
        if value is None:
            # All nullable args have a default (``None`` if not otherwise specified)
            if pa.types.is_struct(self.pyarrow_dtype):
                # Nones are not missing for structs
                return False
            return not self._is_nullable
        return False

    def is_rich_valid(self, value: _TRich):
        """Returns true if value has a valid rich type"""
        try:
            prim = self.from_rich_to_primitive(value, "default_or_error")
            pa.scalar(prim, type=self.pyarrow_dtype)
            return True
        except (TypeError, ValueError):
            return False

    @staticmethod
    def convert_proto_dtype_to_pa_dtype(dtype: pb.ArrowType) -> pa.DataType:
        protobuf_to_unit = {
            pb.TimeUnit.Second: "s",
            pb.TimeUnit.Millisecond: "ms",
            pb.TimeUnit.Microsecond: "us",
            pb.TimeUnit.Nanosecond: "ns",
        }

        if dtype.HasField("NONE"):
            return pa.null()
        elif dtype.HasField("BOOL"):
            return pa.bool_()
        elif dtype.HasField("INT8"):
            return pa.int8()
        elif dtype.HasField("INT16"):
            return pa.int16()
        elif dtype.HasField("INT32"):
            return pa.int32()
        elif dtype.HasField("INT64"):
            return pa.int64()
        elif dtype.HasField("UINT8"):
            return pa.uint8()
        elif dtype.HasField("UINT16"):
            return pa.uint16()
        elif dtype.HasField("UINT32"):
            return pa.uint32()
        elif dtype.HasField("UINT64"):
            return pa.uint64()
        elif dtype.HasField("FLOAT16"):
            return pa.float16()
        elif dtype.HasField("FLOAT32"):
            return pa.float32()
        elif dtype.HasField("FLOAT64"):
            return pa.float64()
        elif dtype.HasField("UTF8"):
            return pa.utf8()
        elif dtype.HasField("LARGE_UTF8"):
            return pa.large_utf8()
        elif dtype.HasField("BINARY"):
            return pa.binary()
        elif dtype.HasField("LARGE_BINARY"):
            return pa.large_binary()
        elif dtype.HasField("DATE32"):
            return pa.date32()
        elif dtype.HasField("DATE64"):
            return pa.date64()
        elif dtype.HasField("TIME32"):
            return pa.time32(protobuf_to_unit[dtype.TIME32])
        elif dtype.HasField("TIME64"):
            return pa.time64(protobuf_to_unit[dtype.TIME64])
        elif dtype.HasField("TIMESTAMP"):
            return pa.timestamp(protobuf_to_unit[dtype.TIMESTAMP.time_unit], tz=dtype.TIMESTAMP.timezone)
        elif dtype.HasField("DURATION"):
            return pa.duration(protobuf_to_unit[dtype.DURATION])
        elif dtype.HasField("DECIMAL"):
            return pa.decimal128(dtype.DECIMAL.precision, dtype.DECIMAL.scale)
        elif dtype.HasField("STRUCT"):
            fields = [
                pa.field(
                    field.name,
                    FeatureConverter.convert_proto_dtype_to_pa_dtype(field.arrow_type),
                    nullable=field.nullable,
                )
                for field in dtype.STRUCT.sub_field_types
            ]
            return pa.struct(fields)
        elif dtype.HasField("LIST"):
            field_type = FeatureConverter.convert_proto_dtype_to_pa_dtype(dtype.LIST.field_type.arrow_type)
            return pa.list_(
                pa.field(
                    dtype.LIST.field_type.name,
                    field_type,
                    nullable=dtype.LIST.field_type.nullable,
                )
            )
        elif dtype.HasField("LARGE_LIST"):
            field_type = FeatureConverter.convert_proto_dtype_to_pa_dtype(dtype.LARGE_LIST.field_type.arrow_type)
            return pa.large_list(
                pa.field(
                    dtype.LARGE_LIST.field_type.name,
                    field_type,
                    nullable=dtype.LARGE_LIST.field_type.nullable,
                )
            )
        elif dtype.HasField("FIXED_SIZE_LIST"):
            field_type = FeatureConverter.convert_proto_dtype_to_pa_dtype(dtype.FIXED_SIZE_LIST.field_type.arrow_type)
            return pa.list_(
                pa.field(
                    dtype.FIXED_SIZE_LIST.field_type.name,
                    field_type,
                    nullable=dtype.FIXED_SIZE_LIST.field_type.nullable,
                ),
                list_size=dtype.FIXED_SIZE_LIST.list_size,
            )

        raise TypeError(f"Could not convert the protobuf type {dtype} to a pyarrow dtype")

    @staticmethod
    def convert_pa_dtype_to_proto_dtype(dtype: pa.DataType) -> pb.ArrowType:
        unit_to_protobuf = {
            "s": pb.TimeUnit.Second,
            "ms": pb.TimeUnit.Millisecond,
            "us": pb.TimeUnit.Microsecond,
            "ns": pb.TimeUnit.Nanosecond,
        }
        if dtype == pa.null():
            return pb.ArrowType(NONE=pb.EmptyMessage())
        elif pa.types.is_boolean(dtype):
            return pb.ArrowType(BOOL=pb.EmptyMessage())
        elif pa.types.is_int8(dtype):
            return pb.ArrowType(INT8=pb.EmptyMessage())
        elif pa.types.is_int16(dtype):
            return pb.ArrowType(INT16=pb.EmptyMessage())
        elif pa.types.is_int32(dtype):
            return pb.ArrowType(INT32=pb.EmptyMessage())
        elif pa.types.is_int64(dtype):
            return pb.ArrowType(INT64=pb.EmptyMessage())
        elif pa.types.is_uint8(dtype):
            return pb.ArrowType(UINT8=pb.EmptyMessage())
        elif pa.types.is_uint16(dtype):
            return pb.ArrowType(UINT16=pb.EmptyMessage())
        elif pa.types.is_uint32(dtype):
            return pb.ArrowType(UINT32=pb.EmptyMessage())
        elif pa.types.is_uint64(dtype):
            return pb.ArrowType(UINT64=pb.EmptyMessage())
        elif pa.types.is_float16(dtype):
            return pb.ArrowType(FLOAT16=pb.EmptyMessage())
        elif pa.types.is_float32(dtype):
            return pb.ArrowType(FLOAT32=pb.EmptyMessage())
        elif pa.types.is_float64(dtype):
            return pb.ArrowType(FLOAT64=pb.EmptyMessage())
        elif pa.types.is_string(dtype):
            return pb.ArrowType(UTF8=pb.EmptyMessage())
        elif pa.types.is_large_string(dtype):
            return pb.ArrowType(LARGE_UTF8=pb.EmptyMessage())
        elif pa.types.is_binary(dtype):
            return pb.ArrowType(BINARY=pb.EmptyMessage())
        elif pa.types.is_large_binary(dtype):
            return pb.ArrowType(LARGE_BINARY=pb.EmptyMessage())
        elif pa.types.is_date32(dtype):
            return pb.ArrowType(DATE32=pb.EmptyMessage())
        elif pa.types.is_date64(dtype):
            return pb.ArrowType(DATE64=pb.EmptyMessage())
        elif dtype == pa.time32("s"):
            return pb.ArrowType(TIME32=pb.TimeUnit.Second)
        elif dtype == pa.time32("ms"):
            return pb.ArrowType(TIME32=pb.TimeUnit.Millisecond)
        elif dtype == pa.time64("us"):
            return pb.ArrowType(TIME64=pb.TimeUnit.Microsecond)
        elif dtype == pa.time64("ns"):
            return pb.ArrowType(TIME64=pb.TimeUnit.Nanosecond)
        elif isinstance(dtype, pa.TimestampType):
            return pb.ArrowType(TIMESTAMP=pb.Timestamp(time_unit=unit_to_protobuf[dtype.unit], timezone=dtype.tz))
        elif isinstance(dtype, pa.DurationType):
            return pb.ArrowType(DURATION=unit_to_protobuf[dtype.unit])
        elif isinstance(dtype, pa.Decimal128Type):
            return pb.ArrowType(DECIMAL=pb.Decimal(precision=dtype.precision, scale=dtype.scale))
        elif isinstance(dtype, pa.Decimal256Type):
            return pb.ArrowType(DECIMAL=pb.Decimal(precision=dtype.precision, scale=dtype.scale))
        elif isinstance(dtype, pa.StructType):
            pb_fields = []
            for pa_field in dtype:
                pa_field = cast(pa.Field, pa_field)
                pb_fields.append(
                    pb.Field(
                        name=pa_field.name,
                        nullable=pa_field.nullable,
                        arrow_type=FeatureConverter.convert_pa_dtype_to_proto_dtype(pa_field.type),
                    )
                )
            return pb.ArrowType(STRUCT=pb.Struct(sub_field_types=pb_fields))
        elif isinstance(dtype, pa.ListType):
            pb_field = pb.Field(
                name=dtype.value_field.name,
                nullable=dtype.value_field.nullable,
                arrow_type=FeatureConverter.convert_pa_dtype_to_proto_dtype(dtype.value_field.type),
            )
            return pb.ArrowType(LIST=pb.List(field_type=pb_field))
        elif isinstance(dtype, pa.LargeListType):
            pb_field = pb.Field(
                name=dtype.value_field.name,
                nullable=dtype.value_field.nullable,
                arrow_type=FeatureConverter.convert_pa_dtype_to_proto_dtype(dtype.value_field.type),
            )
            return pb.ArrowType(LARGE_LIST=pb.List(field_type=pb_field))
        elif isinstance(dtype, pa.FixedSizeListType):
            pb_field = pb.Field(
                name=dtype.value_field.name,
                nullable=dtype.value_field.nullable,
                arrow_type=FeatureConverter.convert_pa_dtype_to_proto_dtype(dtype.value_field.type),
            )
            return pb.ArrowType(FIXED_SIZE_LIST=pb.FixedSizeList(field_type=pb_field, list_size=dtype.list_size))

        raise TypeError(f"Could not convert the pyarrow dtype {dtype} to a protobuf message")
